import logging
import zxing
from tempfile import mkdtemp
import os

from nose import with_setup
from nose.tools import raises

test_barcode_dir = os.path.join(os.path.dirname(__file__), 'barcodes')

test_barcodes = [
    ( 'QR_CODE-easy.png', 'QR_CODE', 'This should be QR_CODE', ),
    ( 'CODE_128-easy.jpg', 'CODE_128', 'This should be CODE_128', ),
    ( 'PDF_417-easy.bmp', 'PDF_417', 'This should be PDF_417', ),
    ( 'AZTEC-easy.jpg', 'AZTEC', 'This should be AZTEC' ),
    ( 'QR CODE (¡filenáme törture test! 😉).png', 'QR_CODE', 'This should be QR_CODE' ),
    ( 'QR_CODE-png-but-wrong-extension.bmp', 'QR_CODE', 'This should be QR_CODE' ),
    ( 'QR_CODE-fun-with-whitespace.png', 'QR_CODE', '\n\r\t\r\r\r\n ' ),
    ( 'QR_CODE-screen_scraping_torture_test.png', 'QR_CODE',
      '\n\\n¡Atención ☹! UTF-8 characters,\n\r embedded newlines,\r &&am&p;& trailing whitespace\t \r ' ),
]

test_reader = None
def setup_reader():
    global test_reader
    if test_reader is None:
        test_reader = zxing.BarCodeReader()


@with_setup(setup_reader)
def test_version():
    global test_reader
    assert test_reader.zxing_version is not None
    assert '.'.join(map(str, test_reader.zxing_version_info)) == test_reader.zxing_version


@with_setup(setup_reader)
def _check_decoding(filename, expected_format, expected_raw, extra={}):
    global test_reader
    path = os.path.join(test_barcode_dir, filename)
    logging.debug('Trying to parse {}, expecting {!r}.'.format(path, expected_raw))
    dec = test_reader.decode(path, pure_barcode=True, **extra)
    if expected_raw is None:
        if dec is not None:
            raise AssertionError('Expected failure, but got result in {} format'.format(expected_format, dec.format))
    else:
        if dec.raw != expected_raw:
            raise AssertionError('Expected {!r} but got {!r}'.format(expected_raw, dec.raw))
        if dec.format != expected_format:
            raise AssertionError('Expected {!r} but got {!r}'.format(expected_format, dec.format))


def test_decoding():
    global test_reader
    yield from ((_check_decoding, filename, expected_format, expected_raw) for filename, expected_format, expected_raw in test_barcodes)


def test_possible_formats():
    yield from ((_check_decoding, filename, expected_format, expected_raw, dict(possible_formats=('CODE_93', expected_format, 'DATA_MATRIX')))
                for filename, expected_format, expected_raw in test_barcodes)


@with_setup(setup_reader)
def test_decoding_multiple():
    reader = zxing.BarCodeReader()
    filenames = [os.path.join(test_barcode_dir, filename) for filename, expected_format, expected_raw in test_barcodes]
    for dec, (filename, expected_format, expected_raw) in zip(reader.decode(filenames, pure_barcode=True), test_barcodes):
        if dec.raw != expected_raw:
            raise AssertionError('{}: Expected {!r} but got {!r}'.format(filename, expected_raw, dec.parsed))
        if dec.format != expected_format:
            raise AssertionError('{}: Expected {!r} but got {!r}'.format(filename, expected_format, dec.format))


def test_parsing():
    dec = zxing.BarCode.parse("""
file:///tmp/default.png (format: FAKE_DATA, type: TEXT):
Raw result:
Élan|\tthe barcode is taking off
Parsed result:
Élan
\tthe barcode is taking off
Found 4 result points:
  Point 0: (24.0,18.0)
  Point 1: (21.0,196.0)
  Point 2: (201.0,198.0)
  Point 3: (205.23952,21.0)
""".encode())
    assert dec.uri == 'file:///tmp/default.png'
    assert dec.format == 'FAKE_DATA'
    assert dec.type == 'TEXT'
    assert dec.raw == 'Élan|\tthe barcode is taking off'
    assert dec.parsed == 'Élan\n\tthe barcode is taking off'
    assert dec.points == [(24.0,18.0),(21.0,196.0),(201.0,198.0),(205.23952,21.0)]


def test_wrong_formats():
    all_test_formats = {fmt for fn,fmt,raw in test_barcodes}
    yield from ((_check_decoding, filename, expected_format, None, dict(possible_formats=all_test_formats - {expected_format}))
                for filename, expected_format, expected_raw in test_barcodes)


@raises(zxing.BarCodeReaderException)
def test_bad_java():
    test_reader = zxing.BarCodeReader(java=os.devnull)
    test_reader.decode(test_barcodes[0][0])


@raises(zxing.BarCodeReaderException)
def test_bad_classpath():
    test_reader = zxing.BarCodeReader(classpath=mkdtemp())
    test_reader.decode(test_barcodes[0][0])


@raises(zxing.BarCodeReaderException)
@with_setup(setup_reader)
def test_nonexistent_file_error():
    global test_reader
    test_reader.decode(os.path.join(test_barcode_dir, 'nonexistent.png'))


@raises(zxing.BarCodeReaderException)
@with_setup(setup_reader)
def test_bad_file_format_error():
    global test_reader
    test_reader.decode(os.path.join(test_barcode_dir, 'bad_format.png'))



########################################################################
#
#  zxing.py -- a quick and dirty wrapper for zxing for python
#
#  this allows you to send images and get back data from the ZXing
#  library:  http://code.google.com/p/zxing/
#

from __future__ import print_function
from urllib.parse import quote
from enum import Enum
import pathlib
import zipfile

from .version import __version__
import subprocess as sp, re, os

class BarCodeReaderException(Exception):
    def __init__(self, message, filename=None, underlying=None):
        self.message, self.filename, self.underlying = message, filename, underlying
        super().__init__(message, filename, underlying)

class BarCodeReader(object):
    cls = "com.google.zxing.client.j2se.CommandLineRunner"

    def __init__(self, classpath=None, java=None):
        self.java = java or 'java'
        self.zxing_version = self.zxing_version_info = None
        if classpath:
            self.classpath = classpath if isinstance(classpath, str) else ':'.join(classpath)
        elif "ZXING_CLASSPATH" in os.environ:
            self.classpath = os.environ.get("ZXING_CLASSPATH","")
        else:
            self.classpath = os.path.join(os.path.dirname(__file__), 'java', '*')

        with zipfile.ZipFile(os.path.join(os.path.dirname(__file__), 'java', 'core.jar')) as c:
            for line in c.open('META-INF/MANIFEST.MF'):
                if line.startswith(b'Bundle-Version: '):
                    self.zxing_version = line.split(b' ', 1)[1].strip().decode()
                    self.zxing_version_info = tuple(int(n) for n in self.zxing_version.split('.'))
                    break

    def decode(self, filenames, try_harder=False, possible_formats=None, pure_barcode=False, products_only=False):
        possible_formats = (possible_formats,) if isinstance(possible_formats, str) else possible_formats

        if isinstance(filenames, str):
            one_file = True
            filenames = filenames,
        else:
            one_file = False
        file_uris = [ pathlib.Path(f).absolute().as_uri() for f in filenames ]

        cmd = [self.java, '-cp', self.classpath, self.cls] + file_uris
        if try_harder:
            cmd.append('--try_harder')
        if pure_barcode:
            cmd.append('--pure_barcode')
        if products_only:
            cmd.append('--products_only')
        if possible_formats:
            for pf in possible_formats:
                cmd += ['--possible_formats', pf ]

        try:
            p = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.STDOUT, universal_newlines=False)
        except FileNotFoundError as e:
            raise BarCodeReaderException("Java binary specified (%s) does not exist" % self.java, self.java, e)
        except PermissionError as e:
            raise BarCodeReaderException("Java binary specified (%s) is not executable" % self.java, self.java, e)
        stdout, stderr = p.communicate()

        if stdout.startswith((b'Error: Could not find or load main class com.google.zxing.client.j2se.CommandLineRunner',
                              b'Exception in thread "main" java.lang.NoClassDefFoundError:')):
            raise BarCodeReaderException("Java JARs not found in classpath (%s)" % self.classpath, self.classpath)
        elif stdout.startswith((b'''Exception in thread "main" javax.imageio.IIOException: Can't get input stream from URL!''',
                                b'''Exception in thread "main" java.util.concurrent.ExecutionException: javax.imageio.IIOException: Can't get input stream from URL!''')):
            raise BarCodeReaderException("Could not find image path: %s" % filenames, filenames)
        elif stdout.startswith(b'''Exception in thread "main" java.io.IOException: Could not load '''):
            raise BarCodeReaderException("Java library could not read image; is it in a supported format?", filenames)
        elif stdout.startswith(b'''Exception '''):
            raise BarCodeReaderException("Unknown Java exception: %s" % stdout)
        elif p.returncode:
            raise BarCodeReaderException("Unexpected Java subprocess return code %d" % p.returncode, self.java)

        if p.returncode:
            codes = [ None for fn in filenames ]
        else:
            file_results = []
            for line in stdout.splitlines(True):
                if line.startswith((b'file:///',b'Exception')):
                    file_results.append(line)
                else:
                    file_results[-1] += line
            codes = [ BarCode.parse(result) for result in file_results ]

        if one_file:
            return codes[0]
        else:
            # zxing (insanely) randomly reorders the output blocks, so we have to put them back in the
            # expected order, based on their URIs
            d = {c.uri: c for c in codes}
            return [d[f] for f in file_uris]


class CLROutputBlock(Enum):
    UNKNOWN = 0
    RAW = 1
    PARSED = 2
    POINTS = 3

class BarCode(object):
    @classmethod
    def parse(cls, zxing_output):
        block = CLROutputBlock.UNKNOWN
        uri = format = type = None
        raw = parsed = b''
        points = []

        for l in zxing_output.splitlines(True):
            if block==CLROutputBlock.UNKNOWN:
                if l.endswith(b': No barcode found\n'):
                    return None
                m = re.match(rb"(\S+) \(format:\s*([^,]+),\s*type:\s*([^)]+)\)", l)
                if m:
                    uri, format, type = m.group(1).decode(), m.group(2).decode(), m.group(3).decode()
                elif l.startswith(b"Raw result:"):
                    block = CLROutputBlock.RAW
            elif block==CLROutputBlock.RAW:
                if l.startswith(b"Parsed result:"):
                    block = CLROutputBlock.PARSED
                else:
                    raw += l
            elif block==CLROutputBlock.PARSED:
                if re.match(rb"Found\s+\d+\s+result\s+points?", l):
                    block = CLROutputBlock.POINTS
                else:
                    parsed += l
            elif block==CLROutputBlock.POINTS:
                m = re.match(rb"\s*Point\s*\d+:\s*\(([\d.]+),([\d.]+)\)", l)
                if m:
                    points.append((float(m.group(1)), float(m.group(2))))

        raw = raw[:-1].decode()
        parsed = parsed[:-1].decode()
        return cls(uri, format, type, raw, parsed, points)

    def __init__(self, uri, format, type, raw, parsed, points):
        self.raw = raw
        self.parsed = parsed
        self.uri = uri
        self.format = format
        self.type = type
        self.points = points

    def __repr__(self):
        return '{}(raw={!r}, parsed={!r}, uri={!r}, format={!r}, type={!r}, points={!r})'.format(
            self.__class__.__name__, self.raw, self.parsed, self.uri, self.format, self.type, self.points)



#!/usr/bin/env python3

from setuptools import setup
from urllib.request import urlretrieve
from urllib.error import URLError
from os import access, path, makedirs, R_OK
import subprocess as sp
import sys, re

if not sys.version_info[0] == 3:
    sys.exit("Python 2.x is not supported; Python 3.x is required.")

########################################

version_py = path.join(path.dirname(__file__), 'zxing', 'version.py')

d = {}
with open(version_py, 'r') as fh:
    exec(fh.read(), d)
    version_pep = d['__version__']

########################################

def download_java_files(force=False):
    files = {'java/javase.jar': 'https://repo1.maven.org/maven2/com/google/zxing/javase/3.4.1/javase-3.4.1.jar',
             'java/core.jar': 'https://repo1.maven.org/maven2/com/google/zxing/core/3.4.1/core-3.4.1.jar',
             'java/jcommander.jar': 'https://repo1.maven.org/maven2/com/beust/jcommander/1.78/jcommander-1.78.jar'}

    for fn, url in files.items():
        p = path.join(path.dirname(__file__), 'zxing', fn)
        d = path.dirname(p)
        if not force and access(p, R_OK):
            print("Already have %s." % p)
        else:
            print("Downloading %s from %s ..." % (p, url))
            try:
                makedirs(d, exist_ok = True)
                urlretrieve(url, p)
            except (OSError, URLError) as e:
                raise
    return list(files.keys())

setup(
    name='zxing',
    version=version_pep,
    description="wrapper for zebra crossing (zxing) barcode library",
    long_description="More information: https://github.com/dlenski/python-zxing",
    url="https://github.com/dlenski/python-zxing",
    author='Daniel Lenski',
    author_email='dlenski@gmail.com',
    packages = ['zxing'],
    package_data = {'zxing': download_java_files()},
    entry_points = {'console_scripts': ['zxing=zxing.__main__:main']},
    test_suite = 'nose.collector',
    license='LGPL v3 or later',
)
